#ifndef __PKMEM_H
#define __PKMEM_H

#include <errno.h>
#include <linux/unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

#define PK_FL_RDLCK			0x000001
#define PK_FL_WRLCK			0x000002

#define EABRT				100
#define EABRTONGOING			101
#define ENOTSTOPPED			102
#define EKMALLOC			103
#define ENOPARAM			104
#define EMPROTECT			105
#define ENOEXIST			106
#define EDELETED			107
#define ENODGRADE			108
#define ENOUGRADE			109
#define EBUG				110
#define EABRTPENDING			111
#define EDENIED				112

static inline void pkerror(const char *cmd)
{
	const int offset = 100; 

	char str[13][128] = { "Process was aborted", 
			      "An abort is going on",
			      "Transaction has not been stopped",
			      "kmalloc() inside the kernel failed",
			      "No parameter by this name",
			      "mprotect() inside the kernel failed",
			      "Doesn't exist",
			      "Deleted",
			      "Cannot downgrade lock",
			      "Cannot upgrade lock",
			      "An internal bug occurred",
			      "An abort is pending",
			      "Denied"
	};
	
	// For err -ENOMEM we can use glibc's error table
	if(errno < offset) 
		perror(cmd);
	else
		printf("%s(): %s\n", cmd, str[errno - offset]);
}

static inline void pkpause(const char *str)
{
	printf("%s\n", str);
	getchar();
}

static inline void *__memcpy(void * dst, void const * src, size_t len)
{
	char * pDst = (char *) dst;
	char const * pSrc = (char const *) src;
	
	while (len--)
		*pDst++ = *pSrc++;
		
	return dst;
}

#define __NR_pk_alloc		294
_syscall1(long, pk_alloc, unsigned long, pages);

#define __NR_pk_alloc_log	295
_syscall1(long, pk_alloc_log, unsigned long, pages);

#define __NR_pk_free		296
_syscall1(long, pk_free, unsigned long, start);

#define __NR_pk_free_log	297
_syscall1(long, pk_free_log, unsigned long, start);

#define __NR_pk_get		298
_syscall2(long, pk_get, unsigned long, start, unsigned long, flags);

#define __NR_pk_end		299
_syscall0(long, pk_end);

#endif
