#include "registry.h"

int *create_registry(const char *regname, int *regfd, int *regsz)
{
	int *reg;
	
	reg = NULL;
	*regfd = shm_open(regname, O_RDWR | O_CREAT, S_IRWXU);
        if(*regfd < 0) { perror("shm_open"); goto out; }

        if(ftruncate(*regfd, *regsz) < 0) { perror("ftruncate"); goto out2; }

        reg = (int *)mmap(0, *regsz, (PROT_READ | PROT_WRITE),
			  MAP_SHARED, *regfd, (off_t) 0);
	if(reg == NULL) { perror("mmap"); goto out2; }
	
	goto out;

out2:
	if(shm_unlink(regname) < 0) { perror("shm_unlink"); }	
out:
	return reg;
}

int destroy_registry(const char *regname, int *reg, int *regsz)
{
	int err;
	
	err = -1;
        if(shm_unlink(regname) < 0) { perror("shm_unlink"); goto out; }
        if(munmap(reg, *regsz) < 0) { perror("munmap"); goto out; }
	err = 0;

out:
	return err;
}

int *map_registry(const char *regname, int *regfd, int *regsz)
{
	int *regptr;
	
	regptr = NULL;
	*regfd = shm_open(regname, O_RDWR | O_CREAT, S_IRWXU);
        if(*regfd < 0) { perror("shm_open"); goto out; }

        if((regptr = (int *)mmap(0, *regsz, (PROT_READ | PROT_WRITE),
				 MAP_SHARED, *regfd, (off_t) 0)) == NULL) {
                perror("mmap");
                goto out2;
        }
	goto out;

out2:
	if(shm_unlink(regname) < 0) { perror("shm_unlink"); }	
out:
	return regptr;
}

int unmap_registry(int *reg, int *regsz)
{
	int err;
	
	err = -1;
        if(munmap(reg, *regsz) < 0) { perror("munmap"); goto out; }
	err = 0;

out:
	return err;
}
