#include "../common/pkmem-user.h"
#include "../common/registry.h"
#include "pc-futex.h"
#include "usersem.h" /* For Futex */

int main(int argc, char *argv[])
{
	char *buffer, *cur;
	const char *filename;
	double tp, seconds;
	int err, fd, sizeof_int, *reg, *regptr, regfd, regsz;
	long long t;
	struct futex *full_futex, *empty_futex;
	unsigned int offset;
	unsigned long start, pages, filesz;

	sizeof_int = sizeof(unsigned int);
	err = 1;
	if(argc != 2) { printf("usage: %s <out-filename>\n", argv[0]); goto out; }
	filename = argv[1];

	// Use the common registry to obtain 
	// <start>, <pages>, <filesize>, <full_futex>, <empty_futex>
	regsz = PK_SHM_SIZE;
	reg = map_registry(PK_SHM_NAME, &regfd, &regsz);
	if(reg == NULL) goto out;
	regptr = reg;
	start = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);

	pages = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);
	filesz = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);
	full_futex = (struct futex *) regptr;
	regptr += sizeof(struct futex);
	empty_futex = (struct futex *) regptr;
	regptr += sizeof(struct futex);

	// Create the buffer of file size
	buffer = malloc(filesz + PAGE_SIZE);
	if(!buffer) { perror("malloc"); goto out; }
	memset(buffer, 0, filesz);
	cur = buffer;

	// Receive the data
	//    - EMPTY? Wait for producer to put something
	//    - COMPLETED? We've received all there is to receive
	// Finally, signal that we've received the data
	t = cpucycles();
	while(1) {
		if(pk_get(start, PK_FL_WRLCK) < 0) { pkerror("pk_get"); goto out; }
		if(GET_STATUS(start) & COMPLETED) {
			dbg("completed");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			break;
		}
		if(GET_STATUS(start) & EMPTY) { 
			dbg("before WAIT");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			if(futex_down(full_futex) != 0) { perror("futex_down"); goto out; }
			dbg("after WAIT");
			continue;
		}
		offset = GET_STATUS(start);
		offset &= ~(EMPTY | FULL | COMPLETED);
		memcpy(cur, (void *)(start + sizeof_int), offset);
		cur += offset;
		SET_STATUS(start, EMPTY);
		dbg("before SIGNAL");
		if(pk_end() < 0) { pkerror("pk_end"); goto out; }
		if(futex_up(empty_futex) < 0) { perror("futex_up"); goto out; }
		dbg("after SIGNAL");
	}

	t = cpucycles() - t;
	if(cur - buffer != filesz) {
		filesz = cur - buffer;
		printf("BUG! Only wrote %lu bytes\n", filesz);
	}

	// Write the data received to file
	fd = open(filename, O_CREAT|O_WRONLY|O_TRUNC, 664);
	if(fd < 0) { perror("open"); goto out; }
	if(write(fd, buffer, filesz) < 0) { close(fd); perror("open"); goto out; }
	if(close(fd) < 0) { perror("close"); goto out; }
	err = 0;

	// Unmap the registry
	if(unmap_registry(reg, &regsz) < 0) goto out;
	
	// Display the stats
	seconds = (double) t / (double) HZ;
	tp = (double) filesz / (double) MEGABYTE;
	tp = tp / seconds;
	printf("%lu\t\t%.*lf\n", pages, 2, tp);

out:
	return err;
}
