#include "../common/pkmem-user.h"
#include "../common/registry.h"
#include "pc-pkfifo.h"

int main(int argc, char *argv[])
{
	unsigned long start, pages, filesz;
	const char *filename;
	char *buffer, *cur;
	unsigned int offset;
	long long t;
	double tp, seconds;
	int err, fd, sizeof_int, *reg, *regptr, regfd, regsz;
	pthread_mutex_t *mutex;
	pthread_cond_t *empty, *full;
	int fullfd, emptyfd, what;

	sizeof_int = sizeof(unsigned int);
	err = 1;
	if(argc != 2) {
		printf("usage: %s <out-filename>\n", argv[0]);
		goto out;
	}
	filename = argv[1];

	// Use the common registry to obtain 
	//  <start>, <pages>, <filesize>
	regsz = PK_SHM_SIZE;
	reg = map_registry(PK_SHM_NAME, &regfd, &regsz);
	if(reg == NULL) goto out;
	regptr = reg;
	start = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);

	pages = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);
	filesz = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);

	// Create the buffer of file size
	buffer = malloc(filesz + PAGE_SIZE);
	if(!buffer) { perror("malloc"); goto out; }
	memset(buffer, 0, filesz);
	cur = buffer;

	// Open the FIFO
	fullfd = open(FULL_FIFO, O_RDWR);
	if(fullfd < 0) { perror("open"); goto out; }
	emptyfd = open(EMPTY_FIFO, O_RDWR);
	if(emptyfd < 0) { perror("open"); goto out; }

	// Receive the data
	//    - EMPTY? Wait for producer to put something
	//    - COMPLETED? We've received all there is to receive
	//   Signal that we've received the data
	t = cpucycles();
	while(1) {
		if(pk_get(start, PK_FL_WRLCK) < 0) { pkerror("pk_get"); goto out; }
		if(GET_STATUS(start) & COMPLETED) {
			dbg("completed");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			break;
		}
		if(GET_STATUS(start) & EMPTY) { 
			dbg("before WAIT");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			if(read(fullfd, &what, sizeof(int)) < 0) { perror("read"); goto out; }
			dbg("after WAIT");
			continue;
		}
		offset = GET_STATUS(start);
		offset &= ~(EMPTY | FULL | COMPLETED);
                volatile char *curptr = (char *)(start + sizeof_int), ch;
                while(curptr - ((char *)(start + sizeof_int)) < offset) {
		  ch = *curptr;
		  curptr++;
                }
		cur += offset;
		SET_STATUS(start, EMPTY);
		dbg("before SIGNAL");
		if(pk_end() < 0) { pkerror("pk_end"); goto out; }
		if(write(emptyfd, &what, sizeof(int)) < 0) { perror("write"); goto out; }
		dbg("after SIGNAL");
	}

	t = cpucycles() - t;
	if(cur - buffer != filesz) {
		filesz = cur - buffer;
		printf("BUG! Only wrote %d bytes\n", filesz);
	}

	// Write the data received to file
	fd = open(filename, O_CREAT|O_WRONLY|O_TRUNC, 664);
	if(fd < 0) { perror("open"); goto out; }
	if(write(fd, buffer, filesz) < 0) { close(fd); perror("open"); goto out; }
	if(close(fd) < 0) { perror("close"); goto out; }
	err = 0;

	// Unmap the registry
	if(unmap_registry(reg, &regsz) < 0) goto out;
	
	// Display the stats
	seconds = (double) t / (double) HZ;
	tp = (double) filesz / (double) MEGABYTE;
	tp = tp / seconds;
	printf("%lu\t\t%.*lf\n", pages, 2, tp);
	
	// Close fifos
	if(close(emptyfd) < 0) { perror("close"); goto out; }
	if(close(fullfd) < 0) { perror("close"); goto out; }
	
out:
	return err;
}
