#include "../common/pkmem-user.h"
#include "pc-pthread.h"

int main(int argc, char *argv[])
{
	char *buffer, *cur;
	const char *filename;
	double tp, seconds;
	int err, fd, sizeof_int, *reg, *regptr, regfd, regsz;
	long long t;
	pthread_cond_t *empty, *full;
	pthread_mutex_t *mutex;
	unsigned int offset;
	unsigned long start, pages, filesz;

	sizeof_int = sizeof(unsigned int);
	err = 1;
	if(argc != 2) {
		printf("usage: %s <out-filename>\n", argv[0]);
		goto out;
	}
	filename = argv[1];

	// Use the common registry to obtain 
	//  <start>, <pages>, <filesize>, <mutex>, <cond>
	regsz = PK_SHM_SIZE;
	reg = (int *)map_registry(PK_SHM_NAME, &regfd, &regsz);
	if(reg == NULL) goto out;
	regptr = reg;
	start = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);

	pages = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);
	filesz = *((unsigned long *) regptr);
	regptr += sizeof(unsigned long);
	mutex = (pthread_mutex_t *) regptr;
	regptr += sizeof(pthread_mutex_t);
	empty = (pthread_cond_t *) regptr;
	regptr += sizeof(pthread_cond_t);
	full = (pthread_cond_t *) regptr;
	regptr += sizeof(pthread_cond_t);

	// Create the buffer of file size
	buffer = malloc(filesz + PAGE_SIZE);
	if(!buffer) { perror("malloc"); goto out; }
	memset(buffer, 0, filesz);
	cur = buffer;

	// Receive the data
	//    - EMPTY? Wait for producer to put something
	//    - COMPLETED? We've received all there is to receive
	//   Signal that we've received the data
	pthread_mutex_lock(mutex);
	t = cpucycles();
	while(1) {
		if(pk_get(start, PK_FL_WRLCK) < 0) { pkerror("pk_get"); goto out; }
		if(GET_STATUS(start) & COMPLETED) {
			dbg("completed");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			break;
		}
		if(GET_STATUS(start) & EMPTY) { 
			dbg("before WAIT");
			if(pk_end() < 0) { pkerror("pk_end"); goto out; }
			pthread_cond_wait(full, mutex);
			dbg("after WAIT");
			continue;
		}
		offset = GET_STATUS(start);
		offset &= ~(EMPTY | FULL | COMPLETED);
		memcpy(cur, (void *)(start + sizeof_int), offset);
		cur += offset;
		SET_STATUS(start, EMPTY);
		dbg("before SIGNAL");
		if(pk_end() < 0) { pkerror("pk_end"); goto out; }
		pthread_cond_signal(empty);
		dbg("after SIGNAL");
	}
	pthread_mutex_unlock(mutex);

	t = cpucycles() - t;
	if(cur - buffer != filesz) {
		filesz = cur - buffer;
		printf("BUG! Only wrote %d bytes\n", filesz);
	}

	// Write the data received to file
	fd = open(filename, O_CREAT|O_WRONLY|O_TRUNC, 664);
	if(fd < 0) { perror("open"); goto out; }
	if(write(fd, buffer, filesz) < 0) { close(fd); perror("open"); goto out; }
	if(close(fd) < 0) { perror("close"); goto out; }
	err = 0;

	// Unmap the registry
	if(unmap_registry(reg, &regsz) < 0) goto out;

	// Display the stats
	seconds = (double) t / (double) HZ;
	tp = (double) filesz / (double) MEGABYTE;
	tp = tp / seconds;
	printf("%lu\t\t%.*lf\n", pages, 2, tp);
out:
	return err;
}
